# service.py

import threading
import warnings
import time
import datetime as dt
from typing import Optional, Union

from represent import represent

Number = Union[int, float]
Host = str
Port = Union[str, int]

__all__ = [
    "ServiceInterface"
]

@represent
class ServiceInterface:
    """The server object to control the communication ith multiple clients."""

    SLEEP = 0.01

    def __init__(self) -> None:
        """Defines the server datasets for clients and client commands."""

        self._timeout_process: Optional[threading.Thread] = None
        self._updating_process: Optional[threading.Thread] = None
        self._refreshing_process: Optional[threading.Thread] = None

        self._blocking = False
        self._updating = False
        self._refreshing = False
    # end __init__

    @property
    def updating(self) -> bool:
        """
        Returns the value of the updating process.

        :return: The updating value.
        """

        return self._updating
    # end updating

    @property
    def blocking(self) -> bool:
        """
        Returns the value of te execution being blocking by the service loop.

        :return: The blocking value.
        """

        return self._blocking
    # end blocking

    @property
    def refreshing(self) -> bool:
        """
        Returns the value of te execution being refreshing by the service loop.

        :return: The refreshing value.
        """

        return self._refreshing
    # end refreshing

    def update(self) -> None:
        """Updates the options according to the screeners."""
    # end update

    def refresh(self) -> None:
        """Updates the options according to the screeners."""
    # end update

    def refreshing_loop(self, refresh: Union[Number, dt.timedelta]) -> None:
        """
        Updates the options according to the screeners.

        :param refresh: The value to refresh the service.
        """

        self._refreshing = True

        if isinstance(refresh, dt.timedelta):
            refresh = refresh.total_seconds()
        # end if

        start = time.time()

        while self.updating:
            current = time.time()

            if (current - start) >= refresh:
                start = current

                self.refresh()
            # end if

            time.sleep(self.SLEEP)
        # end while
    # end update_loop

    def updating_loop(self) -> None:
        """Updates the options according to the screeners."""

        self._updating = True

        while self.updating:
            self.update()

            time.sleep(self.SLEEP)
        # end while
    # end updating_loop

    def blocking_loop(self) -> None:
        """Updates the options according to the screeners."""

        self._blocking = True

        while self.blocking:
            time.sleep(self.SLEEP)
        # end while
    # end blocking_loop

    def start_updating(self) -> None:
        """Starts the updating process."""

        if self.updating:
            warnings.warn(f"Updating process of {self} is already running.")

            return
        # end if

        self._updating_process = threading.Thread(
            target=self.updating_loop
        )

        self._updating_process.start()
    # end start_updating

    def start_refreshing(self, refresh: Union[Number, dt.timedelta]) -> None:
        """
        Starts the refreshing process.

        :param refresh: The value to refresh the service.
        """

        if self.refreshing:
            warnings.warn(f"Refreshing process of {self} is already running.")

            return
        # end if

        self._refreshing_process = threading.Thread(
            target=lambda: self.refreshing_loop(refresh)
        )

        self._refreshing_process.start()
    # end start_refreshing

    def start_blocking(self) -> None:
        """Starts the blocking process."""

        if self.blocking:
            warnings.warn(f"Blocking process of {self} is already running.")

            return
        # end if

        self.blocking_loop()
    # end start_blocking

    @staticmethod
    def start_waiting(
            wait: Union[Number, dt.timedelta, dt.datetime]
    ) -> None:
        """
        Runs a waiting for the process.

        :param wait: The duration of the start_timeout.

        :return: The start_timeout process.
        """

        if isinstance(wait, dt.datetime):
            wait = wait - dt.datetime.now()
        # end if

        if isinstance(wait, dt.timedelta):
            wait = wait.total_seconds()
        # end if

        if isinstance(wait, (int, float)):
            time.sleep(wait)
        # end if
    # end start_waiting

    def run(
            self,
            update: Optional[bool] = False,
            block: Optional[bool] = False,
            refresh: Optional[Union[Number, dt.timedelta]] = None,
            wait: Optional[Union[Number, dt.timedelta, dt.datetime]] = None,
            timeout: Optional[Union[Number, dt.timedelta, dt.datetime]] = None,
    ) -> None:
        """
        Runs the api service.

        :param update: The value to update the service.
        :param block: The value to block the execution and wain for the service.
        :param refresh: The value to refresh the service.
        :param wait: The waiting time.
        :param timeout: The start_timeout for the process.
        """

        if update:
            self.start_updating()
        # end if

        if refresh:
            self.start_refreshing(refresh)
        # end if

        if timeout:
            self.start_timeout(timeout)
        # end if

        if wait:
            self.start_waiting(wait)
        # end if

        if block:
            self.start_blocking()
        # end if
    # end run

    def start_timeout(
            self,
            duration: Union[Number, dt.timedelta, dt.datetime]
    ) -> None:
        """
        Waits to terminate the process.

        :param duration: The amount of seconds to wait before termination.
        """

        if isinstance(duration, dt.datetime):
            duration = duration - dt.datetime.now()
        # end if

        if isinstance(duration, dt.timedelta):
            duration = duration.total_seconds()
        # end if

        self._timeout_process = threading.Thread(
            target=lambda: (time.sleep(duration), self.terminate())
        )

        self._timeout_process.start()
    # end start_timeout

    def stop_blocking(self) -> None:
        """Stops the blocking process."""

        self._blocking = False
    # end stop_blocking

    def stop_updating(self) -> None:
        """Stops the updating process."""

        if self.updating:
            self._updating = False
        # end if

        if (
            isinstance(self._updating_process, threading.Thread) and
            self._updating_process.is_alive()
        ):
            self._updating_process = None
        # end if
    # end stop_updating

    def stop_refreshing(self) -> None:
        """Stops the refreshing process."""

        if self.refreshing:
            self._refreshing = False
        # end if

        if (
            isinstance(self._refreshing_process, threading.Thread) and
            self._refreshing_process.is_alive()
        ):
            self._refreshing_process = None
        # end if
    # end stop_refreshing

    def terminate(self) -> None:
        """Pauses the process of service."""

        self.stop_blocking()
        self.stop_updating()
        self.stop_refreshing()
    # end terminate
# end ServiceInterface


# backend.py

from typing import Any, Optional

import dill

__all__ = [
    "UnSerializableObjectError",
    "UnDeserializableObjectError",
    "SerializationExceptions"
]

SerializationExceptions = (
    TypeError, ValueError, AttributeError,
    dill.PicklingError, dill.PickleError
)

class UnSerializableObjectError(ValueError):
    """A class to represent an exception."""

    def __init__(self, data: Optional[Any] = None) -> None:
        """
        Defines the class attributes.

        :param data: The commands to collect for the exception.
        """

        message = f" {repr(data)} of type {type(data)}" if data is not None else ""

        super().__init__(
            f"Couldn't serialize the object{message}. "
            f"Probably due to the object having weak "
            f"references or C-type pointers."
        )
    # end __init__
# end UnSerializableObjectError

class UnDeserializableObjectError(ValueError):
    """A class to represent an exception."""

    def __init__(self, data: Optional[Any] = None) -> None:
        """
        Defines the class attributes.

        :param data: The commands to collect for the exception.
        """

        message = f" {repr(data)} of type {type(data)}" if data is not None else ""

        super().__init__(
            f"Couldn't deserialize the object{message}. "
            f"Probably due to the object having weak "
            f"references or C-type pointers."
        )
    # end __init__
# end UnDeserializableObjectError


# client.py

import socket
from typing import Optional, Union, Any, Dict

from requests.models import PreparedRequest

from represent import represent

from sockets_communication.bluetooth import bluetooth_socket
from sockets_communication.process import decode
from sockets_communication.communication import Communication

__all__ = [
    "SocketClient",
    "socket_request",
    "bluetooth_socket_request"
]

Socket = socket.socket
Host = str
Port = Union[str, int]

@represent
class SocketClient(Communication):
    """Creates the client to communicate with the server."""

    __slots__ = "connection", "host", "port"

    def __init__(
            self,
            connection: Optional[Socket] = None, *,
            host: Optional[Host] = None,
            port: Optional[Port] = None
    ) -> None:
        """
        Defines the server address and creates the client object.

        :param connection: The connection socket.
        :param host: The ip address of the server.
        :param port: The port for the server connection.
        """

        self.connection: Optional[socket.socket] = connection

        self.host: Optional[str] = host
        self.port: Optional[int] = port
    # end __init__

    def validate_connection(self) -> None:
        """Validates the connection object."""

        if not isinstance(self.connection, socket.socket):
            raise ValueError(
                f"Connection object must be {socket.socket}, "
                f"not: {self.connection}."
            )
        # end if
    # end validate_connection

    def send(self, message: bytes, connection: Optional[Socket] = None) -> None:
        """
        Sends a message to the client or server by its connection.

        :param message: The message to send to the client.
        :param connection: The sockets' connection object.
        """

        if connection is None:
            self.validate_connection()

            connection = self.connection
        # end if

        super().send(message=message, connection=connection)
    # end send

    def receive(self, connection: Optional[Socket] = None) -> bytes:
        """
        Receive a message from the client or server by its connection.

        :param connection: The sockets' connection object.

        :return: The received message from the server.
        """

        if connection is None:
            self.validate_connection()

            connection = self.connection
        # end if

        return super().receive(connection=connection)
    # end receive

    def connect(
            self,
            connection: Optional[Socket] = None, *,
            host: Optional[Host] = None,
            port: Optional[Port] = None
    ) -> None:
        """
        Creates the sockets' connection for the client object with the server.

        :param connection: The connection socket.
        :param host: The ip address of the server.
        :param port: The port for the server connection.
        """

        if connection is not None:
            self.connection = connection
        # end if

        if host is not None:
            self.host = host
        # end if

        if port is not None:
            self.port = port
        # end if

        self.validate_connection()

        self.connection.connect((self.host, self.port))
    # end connect

    def send_message_to_server(self, message: bytes) -> None:
        """
        Sends a message to the server through the sockets' connection.

        :param message: The message to send to the server.
        """

        return self.send(connection=self.connection, message=message)
    # end send_message_to_server

    # defines a method to receive a message form hte server
    def receive_message_from_server(self) -> bytes:
        """
        Gets the received message from the server.

        :return: The received response from the server.
        """

        return self.receive(connection=self.connection)
    # end receive_message_from_server
# end SocketClient

def socket_request(
        connection: Optional[Socket], *,
        host: Host,
        port: Port,
        endpoint: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None
) -> Any:
    """
    Sends a request through the bluetooth sockets.

    :param connection: The connection object.
    :param host: The request host.
    :param endpoint: The path to the endpoint.
    :param port: The sending port.
    :param parameters: The request parameters.

    :return: The returned value.
    """

    client = SocketClient()

    client.connect(connection=connection, host=host, port=port)

    req = PreparedRequest()
    # noinspection HttpUrlsUsage
    req.prepare_url(
        f"http://{host}:{port}/{endpoint or ''}",
        parameters or {}
    )

    client.send(message=req.url.encode())
    content = client.receive()

    client.connection.close()

    return decode(content.decode())
# end socket_request

def bluetooth_socket_request(
        host: Host,
        port: Port,
        endpoint: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None
) -> Any:
    """
    Sends a request through the bluetooth sockets.

    :param host: The request host.
    :param endpoint: The path to the endpoint.
    :param port: The sending port.
    :param parameters: The request parameters.

    :return: The returned value.
    """

    return socket_request(
        connection=bluetooth_socket(), host=host,
        port=port, endpoint=endpoint, parameters=parameters
    )
# end bluetooth_socket_request


# bluetooth.py

import socket

__all__ = [
    "bluetooth_socket"
]

def bluetooth_socket() -> socket.socket:
    """
    Sends a request through the bluetooth sockets.

    :return: The client object..
    """

    return socket.socket(
        socket.AF_BLUETOOTH, socket.SOCK_STREAM,
        socket.BTPROTO_RFCOMM
    )
# end bluetooth_socket